import math
from itertools import chain
from uuid import uuid4

import pandas as pd

from sampo.generator.environment import ContractorGenerationMethod, get_contractor_by_wg
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import DefaultWorkEstimator, WorkTimeEstimator
from sampo.userinput.parser.exception import InputDataException
from sampo.userinput.parser.general_build import add_graph_info, topsort_graph_df, build_work_graph, \
    preprocess_graph_df, break_loops_in_input_graph
from sampo.userinput.parser.history import set_connections_info
from sampo.utilities.name_mapper import NameMapper


class CSVParser:

    @staticmethod
    def read_graph_info(project_info: str | pd.DataFrame,
                        history_data: str | pd.DataFrame,
                        sep_wg: str = ';',
                        sep_history: str = ';',
                        name_mapper: NameMapper | None = None,
                        all_connections: bool = False,
                        change_connections_info: bool = False) -> pd.DataFrame:
        """
        Read the input data about work graph and preprocess it.

        Schema of WorkGraph .csv file:
            mandatory fields:
                activity_id: str - Id of the current task,
                activity_name: str - Name of the current task,
                measurement: str - Measure of the size of the current task (e.g., km, pcs, lit),
                volume: float - Volume of the current task
            optional fields:
                granular_name: str - Task name as in the document,
                predecessor_ids: list[str] - Ids of predecessors of the current task,
                connection_types: list[str] - Types of links between the current task and its predecessors,
                lags: float - Time lags,
                min_req: dict[str, float] - A dictionary containing the minimum amount of each resource
                                            that is required to perform the current task
                max_req: dict[str, float] - A dictionary containing the maximum amount of each resource
                                            that is required to perform the current task
                description: str - A task description
                required_statuses: dict[str, int] - A dictionary containing the zone statuses required
                                                    to perform the current task

        !NOTE! that length of predecessor_ids, connection_types and lags in each cell should be equal
        for the correct predecessor resolution

        Schema of history .csv file (optional data):
            mandatory fields:
                granular_name: str - Task name as in the document,
                first_day: str - Date of commencement of the work,
                last_day: str - Date of completion
                upper_works: list[str] - Names of predecessors of the current task

        ATTENTION!
            1) If you send WorkGraph .csv file without data about connections between tasks, you need to provide .csv
            history file - the SAMPO will be able to reconstruct the connections between tasks based on historical data.
            2) If you send WorkGraph .csv file with column 'predecessor_ids', 'lags' etc. and there is no info in these
            columns, so framework repair the info from history data
            3) If you don't put 'granular_name' and don't receive name_mapper to WorkGraph .csv file, thus you need to
            be sure that activity_name is correct and correlate to 'granular_name' in history .csv file

        :param name_mapper: name mapper that translates 'activity_name' to the name, as from a document
        :param project_info: path to the works' info file
        :param history_data: path to the history data of connection file
        :param change_connections: whether it is necessary to change connections
        :param change_connections_info: whether it is necessary to change connections' information based on history data?
        :param sep_wg: separating character. It's mandatory if you send the WorkGraph .csv
        :param sep_history: separating character. It's mandatory if you send the HistoryData .csv file path
        :return: preprocessed info about works
        """
        graph_df = pd.read_csv(project_info, sep=sep_wg, header=0) if isinstance(project_info,
                                                                                 str) else project_info.copy()
        history_df = pd.read_csv(history_data, sep=sep_history) if isinstance(history_data,
                                                                              str) else history_data.copy()

        if 'predecessor_ids' not in graph_df.columns and history_df.shape[0] == 0:
            raise InputDataException(
                'you have neither history data about tasks nor tasks\' connection info in received .csv file.')

        temp_lst = [math.nan] * graph_df.shape[0]

        for col in ['connection_types', 'lags']:
            if col not in graph_df.columns:
                graph_df[col] = temp_lst
        graph_df = preprocess_graph_df(graph_df, name_mapper)
        works_info = set_connections_info(graph_df, history_df, mapper=name_mapper,
                                          all_connections=all_connections,
                                          change_connections_info=change_connections_info)

        return break_loops_in_input_graph(works_info)

    @staticmethod
    def work_graph_and_contractors(works_info: pd.DataFrame,
                                   contractor_info: str | list[Contractor] | tuple[ContractorGenerationMethod, int]
                                   = (ContractorGenerationMethod.AVG, 1),
                                   contractor_types: list[int] | None = None,
                                   name_mapper: NameMapper | None = None,
                                   work_resource_estimator: WorkTimeEstimator = DefaultWorkEstimator()) \
            -> (WorkGraph, Contractor):
        """
        Gets a info about WorkGraph and Contractors from file .csv.

        Schema of Contractors .csv file (optional data):
            mandatory fields:
                contractor_id: str - Id of the current contractor,
                name: str - Contractor name as in the document
            optional fields:
                {names of resources}: float - each resource is a separate column

        ATTENTION!
            1) If you do not provide work resource estimator, framework uses built-in estimator


        :param works_info: dataFrame that contains preprocessed info about work graph structure
        :param contractor_types:
        :param contractors_number: if we do not receive contractors, we need to know how many contractors the user wants,
        for a generation
        :param contractor_info: path to contractor info .csv file or list of Contractors
        :param work_resource_estimator: work estimator that finds necessary resources, based on the history data
        :return: WorkGraph and list of Contractors
        """

        if contractor_types is None:
            contractor_types = [25]

        # gather resources and contractors based on work resource estimator or contractor info .csv file
        contractors = []
        works_info['activity_name_original'] = works_info.activity_name
        if name_mapper:
            works_info.activity_name = works_info.activity_name.apply(lambda name: name_mapper[name])

        if isinstance(contractor_info, list):
            contractors = contractor_info
        elif isinstance(contractor_info, str):
            # if contractor info is given or contractor info and work resource estimator are received simultaneously
            contractor_df = pd.read_csv(contractor_info, sep=';', header=0) if isinstance(contractor_info,
                                                                                          str) else contractor_info
            for index, row in contractor_df.iterrows():
                specified_id = 'contractor_id' in contractor_df.columns
                column_lag = 2 if specified_id else 1
                contractor_id = row['contractor_id'] if specified_id else str(uuid4())
                contractors.append(
                    Contractor(id=contractor_id,
                               name=row['name'],
                               workers={item[0]: Worker(str(ind), str(item[0]), int(item[1])) for ind, item in
                                        enumerate(row[column_lag:].items())},
                               equipments=dict())
                )

        resources = [dict((worker_req.kind, int(worker_req.volume))
                          for worker_req in work_resource_estimator.find_work_resources(w[0], float(w[1])))
                     for w in works_info.loc[:, ['activity_name', 'volume']].to_numpy()]

        unique_res = list(set(chain(*[r.keys() for r in resources])))
        # works_info.loc[:, unique_res] = DataFrame(resources).fillna(0)

        works_resources = add_graph_info(works_info)
        works_resources = topsort_graph_df(works_resources)
        work_graph = build_work_graph(works_resources, unique_res, work_resource_estimator)

        # if we have no info about contractors or the user send an empty .csv file
        if len(contractors) == 0:
            generation_method, contractors_number = contractor_info
            contractors = [get_contractor_by_wg(work_graph,
                                                method=generation_method,
                                                contractor_id=str(i),
                                                contractor_name='Contractor' + ' ' + str(i + 1))
                           for i in range(contractors_number)]

        return work_graph, contractors



import os

import numpy as np
import torch
import torch.nn.functional as F
from ray.air import session
from torch import nn
from torch.utils.data import DataLoader, TensorDataset

from sampo.scheduler.selection.metrics import one_hot_encode


class NeuralNetType:
    REGRESSION = 'REGRESSION'
    CLASSIFICATION = 'CLASSIFICATION'


class NeuralNet(nn.Module):
    def __init__(self, input_size: int = 13, layer_size: int = 15, layer_count: int = 6, out_size: int = 2,
                 task_type: NeuralNetType = NeuralNetType.REGRESSION):
        super().__init__()

        self.task_type = task_type
        self._layers_count = layer_count
        self._linear0 = torch.nn.Linear(input_size, layer_size)
        self.model = nn.Sequential(self._linear0)
        for i in range(layer_count):
            self.model.add_module(name=f'_relu',
                                  module=torch.nn.ReLU())
            self.model.add_module(name=f'_linear{i + 1}',
                                  module=torch.nn.Linear(layer_size, layer_size))
        self.model.add_module(name=f'_linear{layer_count + 1}',
                              module=torch.nn.Linear(layer_size, out_size))

    def forward(self, X):
        X = self.model(X)
        if self.task_type == NeuralNetType.CLASSIFICATION:
            X = F.softmax(X, dim=0)
        else:
            X = X
        return X


class NeuralNetTrainer:
    def __init__(self, model: NeuralNet, criterion, optimizer, scorer, batch_size):
        self.model = model
        self.criterion = criterion
        self.optimizer = optimizer
        self._device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.scorer = scorer
        self.batch_size = batch_size

    def fit(self, x, y, epochs=10):
        dataset = TensorDataset(x, y)
        loader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

        # Train the model
        for epoch in range(epochs):

            total_loss = 0.0
            total_step = len(loader)

            for i, (image, label) in enumerate(loader):
                image = image.to(self._device)
                label = label.to(self._device)

                outputs = self.model(image)
                loss = self.criterion(outputs, label.float())

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

                total_loss += loss.item()

                if (i + 1) % 100 == 0:
                    print('Epoch [{}/{}], Step [{}/{}], Loss: {:.4f}'
                          .format(epoch + 1, epochs, i + 1, total_step, loss.item()))

            print(f'Epoch [{epoch + 1}/{epochs}], Average Loss: {total_loss / total_step:.4f}')

    def validate(self, x, y):
        dataset = TensorDataset(x, y)
        loader = DataLoader(dataset, batch_size=self.batch_size)

        val_loss = 0.0
        val_score = 0.0
        total = len(loader)

        for i, (image, label) in enumerate(loader):
            with torch.no_grad():
                image = image.to(self._device)
                label = label.to(self._device)

                outputs = self.model(image)
                loss = self.criterion(outputs, label.float())

                val_score += self.scorer(outputs, label.float()).item()

                val_loss += loss.item()

        session.report(
            {'loss': val_loss / total, 'score': val_score / total}
        )
        return val_score / total, val_loss / total

    def predict(self, x: list):
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x.astype(np.float32))
        result = []

        with torch.no_grad():
            for image in x:
                image = image.to(self._device)
                outputs = self.model(image)

                if self.model.task_type is NeuralNetType.CLASSIFICATION:
                    _, predicted = torch.max(outputs.data, 0)
                    result.append(torch.tensor(one_hot_encode(predicted, 2)))
                elif self.model.task_type is NeuralNetType.REGRESSION:
                    result.append(outputs)

        if self.model.task_type is NeuralNetType.CLASSIFICATION:
            return np.asarray([torch.max(v) for v in result])
        elif self.model.task_type is NeuralNetType.REGRESSION:
            return np.asarray([v.numpy() for v in result])

    def predict_proba(self, x: list) -> np.array:
        if isinstance(x, np.ndarray):
            x = torch.from_numpy(x.astype(np.float32))
        result = []
        with torch.no_grad():
            for image in x:
                # image = image.to(self._device)
                outputs = self.model(image)
                result.append(outputs)
        return np.array([v.numpy() for v in result])

    def save_checkpoint(self, tmp_checkpoint_dir, file_name):
        checkpoint_path = os.path.join(tmp_checkpoint_dir, file_name)
        torch.save({
            'model_state_dict': self.model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
        }, checkpoint_path)
        return tmp_checkpoint_dir

    def load_checkpoint(self, tmp_checkpoint_dir):
        checkpoint = torch.load(tmp_checkpoint_dir)
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])



from typing import Optional

import numpy as np
from toposort import toposort_flatten, toposort

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.resource.average_req import AverageReqResourceOptimizer
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator


def get_node_dependencies(wg: WorkGraph) -> dict[str, set[str]]:
    """
    Creates a mapper for nodes in Word Graph that matches each node id to its parents ids
    and leaves only the first node in inseparable chains.

    :param wg: WorkGraph

    :return: dict that maps node id with set of parent nodes ids
    """
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    node_id2inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            node_id2inseparable_parents[child.id] = node.id

    # here we aggregate information about relationships from the whole inseparable chain
    node_id2children = {node.id: set([node_id2inseparable_parents[child.id]
                                      for inseparable in node.get_inseparable_chain_with_self()
                                      for child in inseparable.children]) - {node.id}
                        for node in nodes}

    node_id2parents = {node.id: set() for node in nodes}
    for node_id, children_nodes in node_id2children.items():
        for child_id in children_nodes:
            node_id2parents[child_id].add(node_id)

    return node_id2parents


class TopologicalScheduler(GenericScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order.
    """

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.Topological,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator()):
        super().__init__(scheduler_type=scheduler_type,
                         resource_optimizer=AverageReqResourceOptimizer(),
                         timeline_type=MomentumTimeline,
                         prioritization_f=self._topological_sort,
                         optimize_resources_f=self.get_default_res_opt_function(lambda _: Time(0)),
                         work_estimator=work_estimator)

    # noinspection PyMethodMayBeStatic
    def _topological_sort(self, wg: WorkGraph, work_estimator: WorkTimeEstimator) -> list[GraphNode]:
        """
        Sort 'WorkGraph' in topological order.

        :param wg: WorkGraph
        :param work_estimator: function that calculates execution time of the work
        :return: list of sorted nodes in graph
        """
        dependencies = get_node_dependencies(wg)

        tsorted_nodes_ids = toposort_flatten(dependencies, sort=True)
        tsorted_nodes = [wg[node_id] for node_id in tsorted_nodes_ids]

        return list(reversed(tsorted_nodes))


class RandomizedTopologicalScheduler(TopologicalScheduler):
    """
    Scheduler, that represent 'WorkGraph' in topological order with random.
    """

    def __init__(self,
                 work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                 random_seed: Optional[int] = None):
        super().__init__(work_estimator=work_estimator)
        self._random_state = np.random.RandomState(random_seed)

    def _topological_sort(self, wg: WorkGraph, work_estimator: WorkTimeEstimator) -> list[GraphNode]:
        def shuffle(nodes: set[str]) -> list[str]:
            """
            Shuffle nodes that are on the same level.

            :param nodes: list of nodes
            :return: list of shuffled indices
            """
            nds = list(nodes)
            indices = np.arange(len(nds))
            self._random_state.shuffle(indices)
            return [nds[ind] for ind in indices]

        dependencies = get_node_dependencies(wg)
        tsorted_nodes: list[GraphNode] = [
            wg[node_id] for level in toposort(dependencies)
            for node_id in shuffle(level)
        ]

        return list(reversed(tsorted_nodes))



from typing import TypeVar, Callable, Generic

from sortedcontainers import SortedKeyList

T = TypeVar('T')


class PriorityQueue(Generic[T]):

    _h: SortedKeyList
    _key_getter: Callable[[T], float]

    def __init__(self, lst: list[T], descending: bool = False, key_getter: Callable[[T], float] = lambda x: x):
        comparator = (lambda x: -key_getter(x[1])) if descending else (lambda x: key_getter(x[1]))
        self._h = SortedKeyList([(key_getter(v), v) for v in lst], comparator)
        self._key_getter = key_getter

    @staticmethod
    def empty(descending: bool = False, key_getter: Callable[[T], float] = lambda x: x):
        return PriorityQueue([], descending, key_getter)

    def add(self, value: T):
        self._h.add((self._key_getter(value), value))

    def extract_extremum(self) -> T:
        return self._h.pop()[1]

    def replace(self, old: T, new: T):
        self._h.discard((self._key_getter(old), old))
        self._h.add((self._key_getter(new), new))

    def decrease_key(self, value: T):
        self.replace(value, value)

    def __len__(self):
        return len(self._h)



from typing import Callable

import numpy as np

from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


class IdentityResourceOptimizer(ResourceOptimizer):
    """
    Empty class of resource optimizer.
    """

    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: list[Worker],
                           optimize_array: np.ndarray,
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[list[Worker]], Time]):
        # no actions here
        pass



import math
from operator import itemgetter

from sampo.schemas.exceptions import NotEnoughMaterialsInDepots, NoAvailableResources
from sampo.schemas.landscape import LandscapeConfiguration, MaterialDelivery
from sampo.schemas.resources import Material
from sampo.schemas.sorted_list import ExtendedSortedList
from sampo.schemas.time import Time


class SupplyTimeline:
    def __init__(self, landscape_config: LandscapeConfiguration):
        self._timeline = {}
        self._capacity = {}
        # material -> list of depots, that can supply this type of resource
        self._resource_sources: dict[str, dict[str, int]] = {}
        for landscape in landscape_config.get_all_resources():
            self._timeline[landscape.id] = ExtendedSortedList([(Time(0), landscape.count), (Time.inf(), 0)],
                                                              itemgetter(0))
            self._capacity[landscape.id] = landscape.count
            for count, res in landscape.get_available_resources():
                res_source = self._resource_sources.get(res, None)
                if res_source is None:
                    res_source = {}
                    self._resource_sources[res] = res_source
                res_source[landscape.id] = count

    def can_schedule_at_the_moment(self, id: str, start_time: Time, materials: list[Material], batch_size: int) -> bool:
        return self.find_min_material_time(id, start_time, materials, batch_size) == start_time

    def find_min_material_time(self, id: str, start_time: Time, materials: list[Material], batch_size: int) -> Time:
        sum_materials = sum([material.count for material in materials])
        ratio = sum_materials / batch_size
        batches = max(1, math.ceil(ratio))

        first_batch = [material.copy().with_count(material.count // batches) for material in materials]
        return self.supply_resources(id, start_time, first_batch, True)[1]

    def deliver_materials(self, id: str, start_time: Time, finish_time: Time,
                          materials: list[Material], batch_size: int) -> tuple[list[MaterialDelivery], Time, Time]:
        """
        Models material delivery.

        Delivery performed in batches sized by batch_size.

        :return: pair of material-driven minimum start and finish times
        """
        sum_materials = sum([material.count for material in materials])
        ratio = sum_materials / batch_size
        batches = max(1, math.ceil(ratio))

        first_batch = [material.copy().with_count(material.count // batches) for material in materials]
        other_batches = [first_batch for _ in range(batches - 2)]
        if batches > 1:
            other_batches.append([material.copy().with_count(material.count - batch_material.count * (batches - 1))
                                  for material, batch_material in zip(materials, first_batch)])

        deliveries = []
        d, start_time = self.supply_resources(id, start_time, first_batch, False)
        deliveries.append(d)
        max_finish_time = finish_time
        for batch in other_batches:
            d, finish_time = self.supply_resources(id, max_finish_time, batch, False, start_time)
            deliveries.append(d)
            max_finish_time = finish_time if finish_time > max_finish_time else max_finish_time

        return deliveries, start_time, max_finish_time

    def _find_best_supply(self, material: str, count: int, deadline: Time) -> str:
        # TODO Make better algorithm
        if self._resource_sources.get(material, None) is None:
            raise NoAvailableResources(
                f'Schedule can not be built. No available resource sources with material {material}')
        depots = [depot_id for depot_id, depot_count in self._resource_sources[material].items()
                  if depot_count >= count]
        if not depots:
            raise NotEnoughMaterialsInDepots(
                f"Schedule can not be built. No one supplier has enough '{material}' material")
        depots = [(depot_id, self._timeline[depot_id].bisect_key_left(deadline), -self._capacity[depot_id])
                  for depot_id in depots]
        depots.sort(key=itemgetter(1, 2))

        return depots[0][0]

    def supply_resources(self, work_id: str, deadline: Time, materials: list[Material], simulate: bool,
                         min_supply_start_time: Time = Time(0)) \
            -> tuple[MaterialDelivery, Time]:
        """
        Finds minimal time that given materials can be supplied, greater than given start time

        :param work_id: work id
        :param deadline: the time work starts
        :param materials: material resources that are required to start
        :param simulate: should timeline only find minimum supply time and not change timeline
        :param min_supply_start_time:
        :return: material deliveries, the time when resources are ready
        """
        assert min_supply_start_time <= deadline
        delivery = MaterialDelivery(work_id)
        min_work_start_time = deadline

        def append_in_material_delivery_list(time: Time, count: int, delivery_list: list[tuple[Time, int]]):
            if not simulate:
                if count > need_count:
                    count = need_count
                delivery_list.append((time, count))

        def update_material_timeline_and_res_sources(timeline: ExtendedSortedList, mat_sources: dict[str, int]):
            for time, count in material_delivery_list:
                mat_sources[depot] -= count
                ind = timeline.bisect_key_left(time)
                timeline_time, timeline_count = timeline[ind]
                if timeline_time == time:
                    timeline[ind] = (time, timeline_count - count)
                else:
                    timeline.add((time, capacity - count))

            time, count = timeline[0]
            if not count:
                ind = 1
                is_zero_count = True
                while ind < len(timeline) - 1 and is_zero_count:
                    next_time, next_count = timeline[ind]
                    if not next_count and next_time == time + 1:
                        ind += 1
                        time = next_time
                    else:
                        is_zero_count = False
                if ind == len(timeline) - 1 or timeline[ind][0] != time + 1:
                    ind -= 1
                del timeline[:ind]

        for material in materials:
            if not material.count:
                continue
            material_sources = self._resource_sources[material.name]
            depot = self._find_best_supply(material.name, material.count, deadline)
            material_timeline = self._timeline[depot]
            capacity = self._capacity[depot]
            need_count = material.count
            idx_left = idx_base = material_timeline.bisect_key_right(deadline) - 1
            cur_time = deadline - 1
            material_delivery_list = [] if not simulate else None

            going_right = False

            while need_count > 0:
                # find current period 
                time_left = material_timeline[idx_left][0]
                time_right = material_timeline[idx_left + 1][0]

                if going_right:
                    if cur_time == time_left:
                        time_left_capacity = material_timeline[idx_left][1]
                        if time_left_capacity:
                            append_in_material_delivery_list(cur_time, time_left_capacity, material_delivery_list)
                            need_count -= time_left_capacity
                        cur_time += 1
                    while need_count > 0 and cur_time < time_right:
                        append_in_material_delivery_list(cur_time, capacity, material_delivery_list)
                        need_count -= capacity
                        cur_time += 1
                    if need_count > 0:
                        idx_left += 1
                else:
                    while need_count > 0 and time_left < cur_time and min_supply_start_time <= cur_time:
                        append_in_material_delivery_list(cur_time, capacity, material_delivery_list)
                        need_count -= capacity
                        cur_time -= 1
                    if need_count > 0 and cur_time == time_left and min_supply_start_time <= cur_time:
                        time_left_capacity = material_timeline[idx_left][1]
                        if time_left_capacity:
                            append_in_material_delivery_list(cur_time, time_left_capacity, material_delivery_list)
                            need_count -= time_left_capacity
                        cur_time -= 1
                    if need_count > 0:
                        idx_left -= 1
                        if idx_left < 0 or cur_time < min_supply_start_time:
                            idx_left = idx_base
                            cur_time = deadline
                            going_right = True

            if not simulate:
                update_material_timeline_and_res_sources(material_timeline, material_sources)
                delivery.add_deliveries(material.name, material_delivery_list)

            min_work_start_time = max(min_work_start_time, cur_time)

        return delivery, min_work_start_time

    @property
    def resource_sources(self):
        return self._resource_sources



from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.structurator.prepare_wg_copy import prepare_work_graph_copy, new_start_finish


def delete_graph_node(original_wg: WorkGraph, remove_gn_id: str, change_id: bool = True) -> WorkGraph:
    """
    Deletes a task from WorkGraph.
    If the task consists of several inseparable nodes this function deletes all of those nodes
    :param original_wg: WorkGraph from which a task is deleted
    :param remove_gn_id: id of the node, corresponding to the deleted task.
    If the task consists of several inseparable nodes, this is id of one of them
    :param change_id: do ids in the new graph need to be changed
    :return: new WorkGraph with deleted task
    """
    copied_nodes, old_to_new_ids = prepare_work_graph_copy(original_wg, change_id=change_id)

    copied_remove_gn = copied_nodes[old_to_new_ids[remove_gn_id]]

    inseparable_chain = copied_remove_gn.get_inseparable_chain()
    if inseparable_chain is not None:
        copied_remove_gn = inseparable_chain[len(inseparable_chain) - 1]
        parent_to_delete = copied_remove_gn.inseparable_parent
        while parent_to_delete is not None:
            copied_remove_gn = parent_to_delete
            parent_to_delete = copied_remove_gn.inseparable_parent
            _node_deletion(copied_remove_gn, copied_nodes)
    else:
        _node_deletion(copied_remove_gn, copied_nodes)

    start, finish = new_start_finish(original_wg, copied_nodes, old_to_new_ids)

    return WorkGraph(start, finish)


def _node_deletion(remove_gn: GraphNode, nodes: dict[str, GraphNode]):
    parents = remove_gn.parents
    children = remove_gn.children

    for parent in parents:
        for edge in parent.edges_from:
            if edge.finish.id == remove_gn.id:
                parent.edges_from.remove(edge)

    for child in children:
        for edge in child.edges_to:
            if edge.start.id == remove_gn.id:
                child.edges_to.remove(edge)

    nodes.pop(remove_gn.id)

    for child in children:
        child.add_parents(parents)



import random
import time
from typing import Callable

import numpy as np
from deap import tools
from deap.base import Toolbox

from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome
from sampo.scheduler.genetic.operators import init_toolbox, ChromosomeType, FitnessFunction, TimeFitness
from sampo.scheduler.native_wrapper import NativeWrapper
from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.landscape import LandscapeConfiguration
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator, DefaultWorkEstimator
from sampo.schemas.resources import Worker


def create_toolbox_and_mapping_objects(wg: WorkGraph,
                                       contractors: list[Contractor],
                                       worker_pool: WorkerContractorPool,
                                       population_size: int,
                                       mutate_order: float,
                                       mutate_resources: float,
                                       mutate_zones: float,
                                       init_schedules: dict[
                                           str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                                       rand: random.Random,
                                       spec: ScheduleSpec = ScheduleSpec(),
                                       work_estimator: WorkTimeEstimator = None,
                                       assigned_parent_time: Time = Time(0),
                                       landscape: LandscapeConfiguration = LandscapeConfiguration(),
                                       verbose: bool = True) \
        -> tuple[Toolbox, dict[str, int], dict[int, dict[int, Worker]], dict[int, list[int]]]:
    start = time.time()

    # preparing access-optimized data structures
    nodes = [node for node in wg.nodes if not node.is_inseparable_son()]

    index2node: dict[int, GraphNode] = {index: node for index, node in enumerate(nodes)}
    work_id2index: dict[str, int] = {node.id: index for index, node in index2node.items()}
    worker_name2index = {worker_name: index for index, worker_name in enumerate(worker_pool)}
    index2contractor_obj = {ind: contractor for ind, contractor in enumerate(contractors)}
    index2zone = {ind: zone for ind, zone in enumerate(landscape.zone_config.start_statuses)}
    contractor2index = {contractor.id: ind for ind, contractor in enumerate(contractors)}
    worker_pool_indices = {worker_name2index[worker_name]: {
        contractor2index[contractor_id]: worker for contractor_id, worker in workers_of_type.items()
    } for worker_name, workers_of_type in worker_pool.items()}
    node_indices = list(range(len(nodes)))

    # construct inseparable_child -> inseparable_parent mapping
    inseparable_parents = {}
    for node in nodes:
        for child in node.get_inseparable_chain_with_self():
            inseparable_parents[child] = node

    # here we aggregate information about relationships from the whole inseparable chain
    children = {work_id2index[node.id]: set([work_id2index[inseparable_parents[child].id]
                                             for inseparable in node.get_inseparable_chain_with_self()
                                             for child in inseparable.children])
                for node in nodes}

    parents = {work_id2index[node.id]: set() for node in nodes}
    for node, node_children in children.items():
        for child in node_children:
            parents[child].add(node)

    resources_border = np.zeros((2, len(worker_pool), len(index2node)))
    for work_index, node in index2node.items():
        for req in node.work_unit.worker_reqs:
            worker_index = worker_name2index[req.kind]
            resources_border[0, worker_index, work_index] = req.min_count
            resources_border[1, worker_index, work_index] = req.max_count

    contractor_borders = np.zeros((len(contractor2index), len(worker_name2index)), dtype=int)
    for ind, contractor in enumerate(contractors):
        for ind_worker, worker in enumerate(contractor.workers.values()):
            contractor_borders[ind, ind_worker] = worker.count

    init_chromosomes: dict[str, tuple[ChromosomeType, float, ScheduleSpec]] = \
        {name: (convert_schedule_to_chromosome(work_id2index, worker_name2index,
                                               contractor2index, contractor_borders, schedule, chromosome_spec,
                                               landscape, order),
                importance, chromosome_spec)
         if schedule is not None else None
         for name, (schedule, order, chromosome_spec, importance) in init_schedules.items()}

    if verbose:
        print(f'Genetic optimizing took {(time.time() - start) * 1000} ms')

    return init_toolbox(wg,
                        contractors,
                        worker_pool,
                        landscape,
                        index2node,
                        work_id2index,
                        worker_name2index,
                        index2contractor_obj,
                        index2zone,
                        init_chromosomes,
                        mutate_order,
                        mutate_resources,
                        mutate_zones,
                        landscape.zone_config.statuses.statuses_available(),
                        population_size,
                        rand,
                        spec,
                        worker_pool_indices,
                        contractor2index,
                        contractor_borders,
                        node_indices,
                        parents,
                        children,
                        resources_border,
                        assigned_parent_time,
                        work_estimator), worker_name2index, worker_pool_indices, parents


def build_schedule(wg: WorkGraph,
                   contractors: list[Contractor],
                   worker_pool: WorkerContractorPool,
                   population_size: int,
                   generation_number: int,
                   mutpb_order: float,
                   mutpb_res: float,
                   mutpb_zones: float,
                   init_schedules: dict[str, tuple[Schedule, list[GraphNode] | None, ScheduleSpec, float]],
                   rand: random.Random,
                   spec: ScheduleSpec,
                   landscape: LandscapeConfiguration = LandscapeConfiguration(),
                   fitness_constructor: Callable[
                       [Callable[[list[ChromosomeType]], list[Schedule]]], FitnessFunction] = TimeFitness,
                   work_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                   n_cpu: int = 1,
                   assigned_parent_time: Time = Time(0),
                   timeline: Timeline | None = None,
                   time_border: int = None,
                   optimize_resources: bool = False,
                   deadline: Time = None,
                   verbose: bool = True) \
        -> tuple[ScheduleWorkDict, Time, Timeline, list[GraphNode]]:
    """
    Genetic algorithm.
    Structure of chromosome:
    [[order of job],
     [[numbers of workers types 1 for each job], [numbers of workers types 2], ... ],
      [[border of workers types 1 for each contractor], [border of workers types 2], ...]
    ]

    Different mate and mutation for order and for workers.
    Generate order of job by prioritization from HEFTs and from Topological.
    Generate resources from min to max.
    Overall initial population is valid.

    :return: schedule
    """
    global_start = start = time.time()

    toolbox, *mapping_objects = create_toolbox_and_mapping_objects(wg, contractors, worker_pool, population_size,
                                                                   mutpb_order, mutpb_res, mutpb_zones, init_schedules,
                                                                   rand, spec, work_estimator, assigned_parent_time,
                                                                   landscape, verbose)

    worker_name2index, worker_pool_indices, parents = mapping_objects

    native = NativeWrapper(toolbox, wg, contractors, worker_name2index, worker_pool_indices,
                           parents, work_estimator)
    # create population of a given size
    pop = toolbox.population(n=population_size)

    if verbose:
        print(f'Toolbox initialization & first population took {(time.time() - start) * 1000} ms')

    if native.native:
        native_start = time.time()
        best_chromosome = native.run_genetic(pop, mutpb_order, mutpb_order, mutpb_res, mutpb_res, mutpb_res, mutpb_res,
                                             population_size)
        if verbose:
            print(f'Native evaluated in {(time.time() - native_start) * 1000} ms')
    else:
        have_deadline = deadline is not None
        # save best individuals
        hof = tools.HallOfFame(1, similar=compare_individuals)

        fitness_f = fitness_constructor(native.evaluate) if not have_deadline else TimeFitness(native.evaluate)

        evaluation_start = time.time()

        # map to each individual fitness function
        pop = [ind for ind in pop if toolbox.validate(ind)]
        fitness = fitness_f.evaluate(pop)

        evaluation_time = time.time() - evaluation_start

        for ind, fit in zip(pop, fitness):
            ind.fitness.values = [fit]

        hof.update(pop)
        best_fitness = hof[0].fitness.values[0]

        if verbose:
            print(f'First population evaluation took {evaluation_time * 1000} ms')

        start = time.time()

        generation = 1
        plateau_steps = 0
        new_generation_number = generation_number if not have_deadline else generation_number // 2
        max_plateau_steps = new_generation_number // 2

        while generation <= new_generation_number and plateau_steps < max_plateau_steps \
                and (time_border is None or time.time() - global_start < time_border):
            if verbose:
                print(f'-- Generation {generation}, population={len(pop)}, best fitness={best_fitness} --')

            rand.shuffle(pop)

            offspring = []

            for ind1, ind2 in zip(pop[::2], pop[1::2]):
                # mate
                offspring.extend(toolbox.mate(ind1, ind2, optimize_resources))

            for mutant in offspring:
                # mutation
                if optimize_resources:
                    # resource borders mutation
                    toolbox.mutate_resource_borders(mutant)
                toolbox.mutate(mutant)

            evaluation_start = time.time()

            offspring_fitness = fitness_f.evaluate(offspring)

            for ind, fit in zip(offspring, offspring_fitness):
                ind.fitness.values = [fit]

            evaluation_time += time.time() - evaluation_start

            # renewing population
            pop += offspring
            pop = toolbox.select(pop)
            hof.update([pop[0]])

            prev_best_fitness = best_fitness
            best_fitness = hof[0].fitness.values[0]
            if best_fitness == prev_best_fitness:
                plateau_steps += 1
            else:
                plateau_steps = 0

            if have_deadline and best_fitness <= deadline:
                if all([ind.fitness.values[0] <= deadline for ind in pop]):
                    break

            generation += 1

        # Second stage to optimize resources if deadline is assigned

        if have_deadline:

            fitness_resource = fitness_constructor(native.evaluate)

            if best_fitness > deadline:
                print(f'Deadline not reached !!! Deadline {deadline} < best time {best_fitness}')
                # save best individuals
                hof = tools.HallOfFame(1, similar=compare_individuals)
                pop = [ind for ind in pop if ind.fitness.values[0] == best_fitness]

                evaluation_start = time.time()

                fitness = fitness_resource.evaluate(pop)
                for ind, res_peak in zip(pop, fitness):
                    ind.time = ind.fitness.values[0]
                    ind.fitness.values = [res_peak]

                evaluation_time += time.time() - evaluation_start

                hof.update(pop)
            else:
                optimize_resources = True
                # save best individuals
                hof = tools.HallOfFame(1, similar=compare_individuals)

                pop = [ind for ind in pop if ind.fitness.values[0] <= deadline]

                evaluation_start = time.time()

                fitness = fitness_resource.evaluate(pop)
                for ind, res_peak in zip(pop, fitness):
                    ind.time = ind.fitness.values[0]
                    ind.fitness.values = [res_peak]

                evaluation_time += time.time() - evaluation_start

                hof.update(pop)

                plateau_steps = 0
                new_generation_number = generation_number - generation + 1
                max_plateau_steps = new_generation_number // 2
                best_fitness = hof[0].fitness.values[0]

                if len(pop) < population_size:
                    individuals_to_copy = rand.choices(pop, k=population_size - len(pop))
                    copied_individuals = [toolbox.copy_individual(ind) for ind in individuals_to_copy]
                    for copied_ind, ind in zip(copied_individuals, individuals_to_copy):
                        copied_ind.fitness.values = [ind.fitness.values[0]]
                        copied_ind.time = ind.time
                    pop += copied_individuals

                while generation <= generation_number and plateau_steps < max_plateau_steps \
                        and (time_border is None or time.time() - global_start < time_border):
                    if verbose:
                        print(f'-- Generation {generation}, population={len(pop)}, best peak={best_fitness} --')
                    rand.shuffle(pop)

                    offspring = []

                    for ind1, ind2 in zip(pop[::2], pop[1::2]):
                        # mate
                        offspring.extend(toolbox.mate(ind1, ind2, optimize_resources))

                    for mutant in offspring:
                        # resource borders mutation
                        toolbox.mutate_resource_borders(mutant)
                        # other mutation
                        toolbox.mutate(mutant)

                    evaluation_start = time.time()

                    fitness = fitness_f.evaluate(offspring)

                    for ind, t in zip(offspring, fitness):
                        ind.time = t

                    offspring = [ind for ind in offspring if ind.time <= deadline]

                    fitness_res = fitness_resource.evaluate(offspring)

                    for ind, res_peak in zip(offspring, fitness_res):
                        ind.fitness.values = [res_peak]

                    evaluation_time += time.time() - evaluation_start

                    # renewing population
                    pop += offspring
                    pop = toolbox.select(pop)
                    hof.update([pop[0]])

                    prev_best_fitness = best_fitness
                    best_fitness = hof[0].fitness.values[0]
                    if best_fitness == prev_best_fitness:
                        plateau_steps += 1
                    else:
                        plateau_steps = 0

                    generation += 1

        native.close()

        if verbose:
            print(f'Final time: {best_fitness}')
            print(f'Generations processing took {(time.time() - start) * 1000} ms')
            print(f'Full genetic processing took {(time.time() - global_start) * 1000} ms')
            print(f'Evaluation time: {evaluation_time * 1000}')

        best_chromosome = hof[0]

    scheduled_works, schedule_start_time, timeline, order_nodes = toolbox.chromosome_to_schedule(best_chromosome,
                                                                                                 landscape=landscape,
                                                                                                 timeline=timeline)

    return {node.id: work for node, work in scheduled_works.items()}, schedule_start_time, timeline, order_nodes


def compare_individuals(first: tuple[ChromosomeType], second: tuple[ChromosomeType]):
    return (first[0] == second[0]).all() and (first[1] == second[1]).all() and (first[2] == second[2]).all()



from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field

from sampo.schemas.resources import Worker
from sampo.schemas.time import Time
from sampo.schemas.types import WorkerName
from sampo.schemas.works import WorkUnit


@dataclass
class WorkSpec:
    """
    Here are the container for externally given terms, that the resulting `ScheduledWork` should satisfy.
    Must be used in schedulers.
    :param chain: the chain of works, that should be scheduled one after another, e.g. inseparable,
    that starts from this work. Now unsupported.
    :param assigned_workers: predefined worker team (scheduler should assign this worker team to this work)
    :param assigned_time: predefined work time (scheduler should schedule this work with this execution time)
    :param is_independent: should this work be resource-independent, e.g. executing with no parallel users of
    its types of resources
    """
    chain: list[WorkUnit] | None = None  # TODO Add support
    assigned_workers: dict[WorkerName, int] = field(default_factory=dict)
    assigned_time: Time | None = None
    is_independent: bool = False

@dataclass
class ScheduleSpec:
    """
    Here is the container for externally given terms, that Schedule should satisfy.
    Must be used in schedulers.

    :param work2spec: work specs
    """

    _work2spec: dict[str, WorkSpec] = field(default_factory=lambda: defaultdict(WorkSpec))

    def set_exec_time(self, work: str | WorkUnit, time: Time) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_time = time
        return self

    def assign_workers_dict(self, work: str, workers: dict[WorkerName, int]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = copy(workers)
        return self

    def assign_workers(self, work: str, workers: list[Worker]) -> 'ScheduleSpec':
        if isinstance(work, WorkUnit):
            work = work.id
        self._work2spec[work].assigned_workers = {worker.name: worker.count for worker in workers}
        return self

    def get_work_spec(self, work_id: str) -> WorkSpec:
        return self._work2spec[work_id]
